#include "zeusmp.def"
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              L O R E N T Z _ D            \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine lorentz_d (u1, u2, u3)

c
c  PURPOSE
c    Driver for lorentz update. Updates the velocities due to the
c    transverse magnetic forces (the longitudinal forces are handled 
c    in FORCES).  Adapted from FORCES_D; but LORENTZ needs two slabs 
c    of velocity both above and below
c
c  AUTHOR
c    Mordecai-Mark Mac Low
c
c  LAST MODIFIED
c    3 Mar 98 by M-M Mac Low
c
c  INPUT
c    u[1,2,3]    velocity components before acceleration
c
c  OUTPUT
c    u[1,2,3]    velocity components after update
c
c  EXTERNALS:
c    BVALV1  , BVALV2  , BVALV3
c    BVALD   , BVALE  
c    LORENTZ
c
c.......................................................................
c
       implicit NONE
#include "mod_names.h"
#include "param.h"
#include "grid.h"
#include "field.h"

       REAL u1(in,jn,kn), u2(in,jn,kn), u3(in,jn,kn)


#ifndef NO_FORCES

#include "root.h"

#include "bndry.h"
#include "scratch.h"
#include "mpicom.h"

#ifdef PROFILING
#ifdef MPE_USED
#else
#include "clocks.h"
#endif /* MPE_USED */
#endif /* PROFILING */
c
       integer i, k1, k2
c
c      External statements
c
       external      lorentz
     1             , bvalv1  , bvalv2  , bvalv3  , bvale   , bvald

c
c-----------------------------------------------------------------------
c
c Divide the computational volume into three equal pieces.  We must
c have at least 5 active zones in the 3-direction.
c
       k1 = int( real( ke - ks + 1 ) / 3.0 ) + ks
       k2 = int( real( ke - ks + 1 ) / 3.0 ) + k1
c
c
#ifdef DEBUG
       write(*,"(/'LORENTZ_D: ',i2,' calling LORENTZ.')") myid
#endif /* DEBUG */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal1 = real(wclock)
       cpu1 = real(cputime)
#endif /* MPE_USED */
#endif /* PROFILING */

c......................................................................
c
c i boundaries
c
c    1) Post sends and receives.  We use d(i-2:i+1) in LORENTZ
c       so we need to pass only 1 "p" layer of density boundary
c       values, along with 2 "m" layers.
  
c       The fields should have been fully updated in CT.  The velocities,
c       on the other hand, need 2 layers.  Therefore, the "boundary" slices
c       that are called separately are 2 layers thick rather than 1 as in 
c       forces_d.
c
c       By exchanging the i, j, and k boundary data in three separate 
c       steps, we ensure that the corner and edge data are correctly
c       updated.
c
       nreq = 0
       nsub = nsub + 1
       call bvald  (3,1,0,0,0,0,d )

       call bvalv1 (3,3,0,0,0,0,v1)
       call bvalv2 (3,3,0,0,0,0,v2)
       call bvalv3 (3,3,0,0,0,0,v3)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,1,2,2) = timers(nsubstps,1,2,2) + wal1  - wal0
       timers(nsubstps,1,2,1) = timers(nsubstps,1,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do first portion of the interior points.
c
       call lorentz (is+2,ie-2,js+2,je-2,ks+2,k1,v1,v2,v3,u1,u2,u3)
c
c       subroutine lorentz (ibeg,iend,jbeg,jend,kbeg,kend
c     &                   ,u1,u2,u3,w1,w2,w3)
cc
cc Arrays u1 , u2 , u3  hold the old velocity values, while
cc        w1 , w2 , w3  receive the updated values.
c

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,1,2,2) = timers(1,1,2,2) + wal1  - wal0
       timers(1,1,2,1) = timers(1,1,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,1,1,2) = timers(1,1,1,2) + wal1  - wal0
       timers(1,1,1,1) = timers(1,1,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif /* MPI_USED */
c......................................................................
c
c j boundaries
c
c    1) Post sends and receives. 
c
       nreq = 0
       nsub = nsub + 1

       call bvald  (0,0,3,1,0,0,d )

       call bvalv1 (0,0,3,3,0,0,v1)
       call bvalv2 (0,0,3,3,0,0,v2)
       call bvalv3 (0,0,3,3,0,0,v3)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,2,2,2) = timers(nsubstps,2,2,2) + wal1  - wal0
       timers(nsubstps,2,2,1) = timers(nsubstps,2,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do middle portion of the interior points, plus some on borders.
c
       call lorentz (is  ,is+1,js+2,je-2,ks+1,k1,v1,v2,v3,u1,u2,u3)
       call lorentz (ie-1,ie  ,js+2,je-2,ks+1,k1,v1,v2,v3,u1,u2,u3)
       call lorentz (is  ,ie  ,js+2,je-2,k1+1,k2,v1,v2,v3,u1,u2,u3)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,2,2,2) = timers(1,2,2,2) + wal1  - wal0
       timers(1,2,2,1) = timers(1,2,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,2,1,2) = timers(1,2,1,2) + wal1  - wal0
       timers(1,2,1,1) = timers(1,2,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif /* MPI_USED */
c......................................................................
c
c k boundaries
c
c    1) Post sends and receives.  
c
       nreq = 0
       nsub = nsub + 1
       call bvald  (0,0,0,0,3,1,d )
       call bvalv1 (0,0,0,0,3,3,v1)
       call bvalv2 (0,0,0,0,3,3,v2)
       call bvalv3 (0,0,0,0,3,3,v3)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,3,2,2) = timers(nsubstps,3,2,2) + wal1  - wal0
       timers(nsubstps,3,2,1) = timers(nsubstps,3,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do last portion of the interior points, plus some on borders.
c
       call lorentz (is  ,ie  ,js  ,js+1,ks+2,k2  ,v1,v2,v3,u1,u2,u3)
       call lorentz (is  ,ie  ,je-1,je  ,ks+2,k2  ,v1,v2,v3,u1,u2,u3)
       call lorentz (is  ,ie  ,js  ,je  ,k2+1,ke-2,v1,v2,v3,u1,u2,u3)
c
c      Mark the velocity boundary values out of date.
c
       do 10 i = 1,6
         bvstat(i,3) = 0      !  v1
         bvstat(i,4) = 0      !  v2
         bvstat(i,5) = 0      !  v3
10     continue

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,3,2,2) = timers(1,3,2,2) + wal1  - wal0
       timers(1,3,2,1) = timers(1,3,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete. 
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,3,1,2) = timers(1,3,1,2) + wal1  - wal0
       timers(1,3,1,1) = timers(1,3,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif /* MPI_USED */
c......................................................................
c
c Finally, do the remaining border zones.
c
       call lorentz (is  ,ie  ,js  ,je  ,ks  ,ks+1,v1,v2,v3,u1,u2,u3)  
       call lorentz (is  ,ie  ,js  ,je  ,ke-1,ke  ,v1,v2,v3,u1,u2,u3)  

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,4,2,2) = timers(1,4,2,2) + wal1  - wal0
       timers(1,4,2,1) = timers(1,4,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#else  /* not NO_FORCES */
       integer i,j,k
c
c-----------------------------------------------------------------------
c
c Just copy the old velocities to the new for all active zones.
c
       do k=ks,ke
         do j=js,je
           do i=is,ie
             u1(i,j,k) = v1(i,j,k)
             u2(i,j,k) = v2(i,j,k)
             u3(i,j,k) = v3(i,j,k)
           enddo ! i
         enddo ! j
       enddo ! k
#endif /* not NO_FORCES */

       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////           L O R E N T Z _ D               \\\\\\\\\\
c
c=======================================================================
c
